from collections.abc import Iterable
from typing import Any
import pytest
from anys import ANY_INT, ANY_STR, any_in
from test_lib import assert_equal, assert_not_equal


@pytest.mark.parametrize(
    "seq,value",
    [
        ([0, 1, 2], 0),
        ([0, 1, 2], 1),
        ([0, 1, 2], 2),
        ("abc", "a"),
        ([ANY_STR, ANY_INT], "foo"),
        ([ANY_STR, ANY_INT], 42),
        ([ANY_INT, "π"], 42),
        ([ANY_INT, "π"], "π"),
        (("foo", "bar"), "foo"),
        ({"foo", "bar"}, "foo"),
        ({"foo": 42, "bar": 23}, "foo"),
        (b"abc", 97),
    ],
)
def test_from_any_in_eq(seq: Iterable, value: Any) -> None:
    assert_equal(any_in(seq), value)
    assert repr(any_in(seq)) == f"AnyIn({list(seq)})"


@pytest.mark.parametrize(
    "seq,value",
    [
        ([0, 1, 2], 3),
        ("abc", "ab"),
        ([ANY_STR, ANY_INT], 3.14),
        ([ANY_INT, "π"], "pi"),
        (("foo", "bar"), "quux"),
        ({"foo", "bar"}, "quux"),
        ({"foo": 42, "bar": 23}, 42),
        (b"abc", 65),
    ],
)
def test_from_any_in_neq(seq: Iterable, value: Any) -> None:
    assert_not_equal(any_in(seq), value)



from collections.abc import Mapping
from types import SimpleNamespace
from typing import Any
import pytest
from anys import ANY_INT, ANY_STR, any_with_attrs
from test_lib import assert_equal, assert_not_equal


@pytest.mark.parametrize(
    "arg,value",
    [
        ({"foo": ANY_INT, "bar": ANY_STR}, SimpleNamespace(foo=42, bar="glarch")),
        (
            {"foo": ANY_INT, "bar": ANY_STR},
            SimpleNamespace(foo=42, bar="glarch", quux=3.14),
        ),
        ({}, SimpleNamespace()),
        ({}, SimpleNamespace(foo=42, bar="glarch")),
        ({"foo": ANY_INT, "bar": "glarch"}, SimpleNamespace(foo=42, bar="glarch")),
        ({"foo": 42, "bar": "glarch"}, SimpleNamespace(foo=42, bar="glarch")),
    ],
)
def test_any_with_attrs_eq(arg: Mapping, value: Any) -> None:
    assert_equal(any_with_attrs(arg), value)
    assert repr(any_with_attrs(arg)) == f"AnyWithAttrs({arg!r})"


@pytest.mark.parametrize(
    "arg,value",
    [
        ({"foo": ANY_INT, "bar": ANY_STR}, SimpleNamespace(foo=42)),
        ({"foo": ANY_INT, "bar": ANY_STR}, SimpleNamespace(foo=42, bar=23)),
        ({"foo": ANY_INT, "bar": ANY_STR}, SimpleNamespace()),
        ({"foo": 42, "bar": "glarch"}, SimpleNamespace()),
        ({"foo": ANY_INT, "bar": ANY_STR}, 42),
        ({1: ANY_INT}, [42, 23]),
        ({"foo": ANY_INT, "bar": ANY_STR}, {"foo": 42, "bar": "glarch"}),
    ],
)
def test_any_with_attrs_neq(arg: Mapping, value: Any) -> None:
    assert_not_equal(any_with_attrs(arg), value)



from functools import partial
from operator import lt
from typing import Any, Callable
import pytest
from anys import any_func
from test_lib import assert_equal, assert_not_equal


@pytest.mark.parametrize(
    "func,value",
    [
        (lambda x: x < 5, 4),
        (lambda x, y="foo": x in y, "oo"),
        (callable, callable),
        (len, "foo"),
        (len, [1, 2, 3]),
        (partial(lt, 5), 42),
    ],
)
def test_any_func_eq(func: Callable, value: Any) -> None:
    assert_equal(any_func(func), value)
    assert repr(any_func(func)) == f"AnyFunc({func!r})"


@pytest.mark.parametrize(
    "func,value",
    [
        (lambda x: x < 5, 42),
        (lambda x, y="foo": x in y, "bar"),
        (callable, 42),
        (len, ""),
        (len, []),
        (partial(lt, 5), -5),
        (len, 42),
        (lambda x, y="foo": x in y, 42),
        (partial(lt, 5), "foo"),
    ],
)
def test_any_func_neq(func: Callable, value: Any) -> None:
    assert_not_equal(any_func(func), value)
    assert repr(any_func(func)) == f"AnyFunc({func!r})"



from typing import Any, AnyStr, Union
import pytest
from anys import any_substr
from test_lib import assert_equal, assert_not_equal


@pytest.mark.parametrize(
    "s,sub",
    [
        ("abc", "a"),
        ("abc", "ab"),
        ("abc", "bc"),
        (b"abc", b"a"),
        (b"abc", b"ab"),
        (b"abc", b"bc"),
        (b"abc", 97),
    ],
)
def test_any_substr_eq(s: AnyStr, sub: Union[AnyStr, int]) -> None:
    assert_equal(any_substr(s), sub)
    assert repr(any_substr(s)) == f"AnySubstr({s!r})"


@pytest.mark.parametrize(
    "s,sub",
    [
        ("abc", 97),
        (b"abc", 42),
        ("abc", b"abc"),
        (b"abc", "abc"),
        ("π", "pi"),
    ],
)
def test_any_substr_neq(s: AnyStr, sub: Any) -> None:
    assert_not_equal(any_substr(s), sub)
    assert repr(any_substr(s)) == f"AnySubstr({s!r})"



from typing import Any
import pytest
from anys import ANY_FALSY, ANY_TRUTHY
from test_lib import assert_equal, assert_not_equal


@pytest.mark.parametrize(
    "value",
    [
        True,
        1,
        3.14,
        1j,
        "foo",
        {"foo": 42},
        [1, 2, 3],
        {1, 2, 3},
        (1, 2, 3),
        object(),
    ],
)
def test_any_truthy_eq_any_falsy_neq(value: Any) -> None:
    assert_equal(ANY_TRUTHY, value)
    assert_not_equal(ANY_FALSY, value)


@pytest.mark.parametrize(
    "value",
    [
        None,
        False,
        0,
        0.0,
        0j,
        "",
        {},
        [],
        set(),
        (),
    ],
)
def test_any_truthy_neq_any_falsy_eq(value: Any) -> None:
    assert_not_equal(ANY_TRUTHY, value)
    assert_equal(ANY_FALSY, value)


def test_any_truthy_repr() -> None:
    assert repr(ANY_TRUTHY) == "ANY_TRUTHY"


def test_any_falsy_repr() -> None:
    assert repr(ANY_FALSY) == "ANY_FALSY"



from collections.abc import Mapping
from types import SimpleNamespace
from typing import Any
import pytest
from anys import ANY_INT, ANY_STR, any_with_entries
from test_lib import assert_equal, assert_not_equal


@pytest.mark.parametrize(
    "arg,value",
    [
        ({"foo": ANY_INT, "bar": ANY_STR}, {"foo": 42, "bar": "glarch"}),
        ({"foo": ANY_INT, "bar": ANY_STR}, {"foo": 42, "bar": "glarch", "quux": 3.14}),
        ({}, {}),
        ({}, {"foo": 42, "bar": "glarch"}),
        ({"foo": ANY_INT, "bar": "glarch"}, {"foo": 42, "bar": "glarch"}),
        ({"foo": 42, "bar": "glarch"}, {"foo": 42, "bar": "glarch"}),
        ({1: ANY_INT, 2: ANY_STR}, [3.14, 42, "glarch"]),
    ],
)
def test_any_with_entries_eq(arg: Mapping, value: Any) -> None:
    assert_equal(any_with_entries(arg), value)
    assert repr(any_with_entries(arg)) == f"AnyWithEntries({arg!r})"


@pytest.mark.parametrize(
    "arg,value",
    [
        ({"foo": ANY_INT, "bar": ANY_STR}, {"foo": 42}),
        ({"foo": ANY_INT, "bar": ANY_STR}, {"foo": 42, "bar": 23}),
        ({"foo": ANY_INT, "bar": ANY_STR}, {}),
        ({"foo": 42, "bar": "glarch"}, {}),
        ({"foo": ANY_INT, "bar": ANY_STR}, 42),
        ({"1": ANY_INT}, [42, 23]),
        ({1: ANY_STR}, [42, 23]),
        ({5: ANY_STR}, [42, 23]),
        ({"foo": ANY_INT, "bar": ANY_STR}, SimpleNamespace(foo=42, bar="glarch")),
    ],
)
def test_any_with_entries_neq(arg: Mapping, value: Any) -> None:
    assert_not_equal(any_with_entries(arg), value)



from typing import Any


def assert_equal(a: Any, value: Any) -> None:
    assert a == value
    assert value == a
    assert {"foo": value} == {"foo": a}
    assert {"foo": a} == {"foo": value}
    assert [1, 2, value, 3] == [1, 2, a, 3]
    assert [1, 2, a, 3] == [1, 2, value, 3]


def assert_not_equal(a: Any, value: Any) -> None:
    assert a != value
    assert value != a
    assert {"foo": value} != {"foo": a}
    assert {"foo": a} != {"foo": value}
    assert [1, 2, value, 3] != [1, 2, a, 3]
    assert [1, 2, a, 3] != [1, 2, value, 3]



from typing import Any
import pytest
from anys import ANY_FLOAT, ANY_INT, ANY_STR, any_contains
from test_lib import assert_equal, assert_not_equal


@pytest.mark.parametrize(
    "seq,value",
    [
        ([0, 1, 2], 0),
        ([0, 1, 2], 1),
        ([0, 1, 2], 2),
        ("abc", "a"),
        ("abc", "ab"),
        (["foo", 42], "foo"),
        (["foo", 42], 42),
        (["foo", 42], ANY_STR),
        (["foo", 42], ANY_INT),
        (("foo", "bar"), "foo"),
        ({"foo", "bar"}, "foo"),
        ({"foo": 42, "bar": 23}, "foo"),
        (b"abc", 97),
    ],
)
def test_from_any_contains_eq(seq: Any, value: Any) -> None:
    assert_equal(any_contains(value), seq)
    assert repr(any_contains(value)) == f"AnyContains({value!r})"


@pytest.mark.parametrize(
    "seq,value",
    [
        ([0, 1, 2], 3),
        (["foo", 42], 3.14),
        (["foo", 42], ANY_FLOAT),
        (["foo", "bar"], ANY_INT),
        ([42, 23], ANY_STR),
        (("foo", "bar"), "quux"),
        ({"foo", "bar"}, "quux"),
        ({"foo": 42, "bar": 23}, 42),
        (b"abc", 65),
        (42, 42),
    ],
)
def test_from_any_contains_neq(seq: Any, value: Any) -> None:
    assert_not_equal(any_contains(value), seq)



import re
from typing import Any, AnyStr, Pattern, Union
import pytest
from anys import any_fullmatch, any_match, any_search
from test_lib import assert_equal, assert_not_equal


@pytest.mark.parametrize(
    "rgx,value",
    [
        (r"\d+", "12345"),
        (r"\d+", "123abc"),
        (re.compile(r"\d+"), "12345"),
        (re.compile(r"\d+"), "123abc"),
        (br"\d+", b"12345"),
        (br"\d+", b"123abc"),
        (re.compile(br"\d+"), b"12345"),
        (re.compile(br"\d+"), b"123abc"),
    ],
)
def test_any_match_eq(rgx: Union[AnyStr, Pattern[AnyStr]], value: AnyStr) -> None:
    assert_equal(any_match(rgx), value)
    assert repr(any_match(rgx)) == f"AnyMatch({rgx!r})"


@pytest.mark.parametrize(
    "rgx,value",
    [
        (r"\d+", b"12345"),
        (r"\d+", b"123abc"),
        (re.compile(r"\d+"), b"12345"),
        (re.compile(r"\d+"), b"123abc"),
        (br"\d+", "12345"),
        (br"\d+", "123abc"),
        (re.compile(br"\d+"), "12345"),
        (re.compile(br"\d+"), "123abc"),
        (r"\d+", "abc123"),
        (br"\d+", b"abc123"),
        (r"\d+", 42),
        (re.compile(r"\d+"), 42),
    ],
)
def test_any_match_neq(rgx: Union[AnyStr, Pattern[AnyStr]], value: Any) -> None:
    assert_not_equal(any_match(rgx), value)


@pytest.mark.parametrize(
    "rgx,value",
    [
        (r"\d+", "12345"),
        (r"\d+", "123abc"),
        (re.compile(r"\d+"), "12345"),
        (re.compile(r"\d+"), "123abc"),
        (br"\d+", b"12345"),
        (br"\d+", b"123abc"),
        (re.compile(br"\d+"), b"12345"),
        (re.compile(br"\d+"), b"123abc"),
        (r"\d+", "abc123"),
        (br"\d+", b"abc123"),
    ],
)
def test_any_search_eq(rgx: Union[AnyStr, Pattern[AnyStr]], value: AnyStr) -> None:
    assert_equal(any_search(rgx), value)
    assert repr(any_search(rgx)) == f"AnySearch({rgx!r})"


@pytest.mark.parametrize(
    "rgx,value",
    [
        (r"\d+", b"12345"),
        (r"\d+", b"123abc"),
        (re.compile(r"\d+"), b"12345"),
        (re.compile(r"\d+"), b"123abc"),
        (br"\d+", "12345"),
        (br"\d+", "123abc"),
        (re.compile(br"\d+"), "12345"),
        (re.compile(br"\d+"), "123abc"),
        (r"\d+", 42),
        (re.compile(r"\d+"), 42),
    ],
)
def test_any_search_neq(rgx: Union[AnyStr, Pattern[AnyStr]], value: Any) -> None:
    assert_not_equal(any_search(rgx), value)


@pytest.mark.parametrize(
    "rgx,value",
    [
        (r"\d+", "12345"),
        (re.compile(r"\d+"), "12345"),
        (br"\d+", b"12345"),
        (re.compile(br"\d+"), b"12345"),
    ],
)
def test_any_fullmatch_eq(rgx: Union[AnyStr, Pattern[AnyStr]], value: AnyStr) -> None:
    assert_equal(any_fullmatch(rgx), value)
    assert repr(any_fullmatch(rgx)) == f"AnyFullmatch({rgx!r})"


@pytest.mark.parametrize(
    "rgx,value",
    [
        (r"\d+", "123abc"),
        (br"\d+", b"123abc"),
        (r"\d+", b"12345"),
        (r"\d+", b"123abc"),
        (re.compile(r"\d+"), b"12345"),
        (re.compile(r"\d+"), b"123abc"),
        (br"\d+", "12345"),
        (br"\d+", "123abc"),
        (re.compile(br"\d+"), "12345"),
        (re.compile(br"\d+"), "123abc"),
        (r"\d+", "abc123"),
        (br"\d+", b"abc123"),
        (r"\d+", 42),
        (re.compile(r"\d+"), 42),
    ],
)
def test_any_fullmatch_neq(rgx: Union[AnyStr, Pattern[AnyStr]], value: Any) -> None:
    assert_not_equal(any_fullmatch(rgx), value)



from typing import Any, Union
import pytest
from anys import (
    ANY_FLOAT,
    ANY_INT,
    ANY_LIST,
    ANY_STR,
    AnyAnd,
    AnyOr,
    any_func,
    any_gt,
    any_lt,
)
from test_lib import assert_equal, assert_not_equal


@pytest.mark.parametrize("value", [-23, 0, 42, True, False, "", "foo"])
def test_any_int_or_str_eq(value: Union[int, str]) -> None:
    assert_equal(ANY_INT | ANY_STR, value)


@pytest.mark.parametrize("value", [None, (), 3.14, 1 + 2j, 1.2 + 3.4j, [], {}, b"bar"])
def test_any_int_or_str_neq(value: Any) -> None:
    assert_not_equal(ANY_INT | ANY_STR, value)


def test_any_int_or_str_repr() -> None:
    assert repr(ANY_INT | ANY_STR) == "AnyOr(ANY_INT, ANY_STR)"


def test_or() -> None:
    a = ANY_INT | ANY_STR
    assert isinstance(a, AnyOr)
    assert len(a.args) == 2
    assert a.args[0] is ANY_INT
    assert a.args[1] is ANY_STR


def test_or3() -> None:
    a = ANY_INT | ANY_STR | ANY_FLOAT
    assert isinstance(a, AnyOr)
    assert len(a.args) == 3
    assert a.args[0] is ANY_INT
    assert a.args[1] is ANY_STR
    assert a.args[2] is ANY_FLOAT


def test_or_or() -> None:
    a1 = ANY_INT | ANY_FLOAT
    a2 = ANY_STR | ANY_LIST
    a = a1 | a2
    assert len(a.args) == 4
    assert a.args[0] is ANY_INT
    assert a.args[1] is ANY_FLOAT
    assert a.args[2] is ANY_STR
    assert a.args[3] is ANY_LIST


@pytest.mark.parametrize("value", [24, 30, 32.5, 41])
def test_any_gt_and_any_lt_eq(value: Union[int, float]) -> None:
    assert_equal(any_gt(23) & any_lt(42), value)


@pytest.mark.parametrize(
    "value", [22, 23, 42, 43, None, (), 3.14, 1 + 2j, 1.2 + 3.4j, [], {}, "foo", b"bar"]
)
def test_any_gt_and_any_lt_neq(value: Any) -> None:
    assert_not_equal(any_gt(23) & any_lt(42), value)


def test_any_gt_and_any_lt_repr() -> None:
    assert repr(any_gt(23) & any_lt(42)) == "AnyAnd(AnyGT(23), AnyLT(42))"


def test_and() -> None:
    a1 = any_gt(23)
    a2 = any_lt(42)
    a = a1 & a2
    assert isinstance(a, AnyAnd)
    assert len(a.args) == 2
    assert a.args[0] is a1
    assert a.args[1] is a2


def test_and3() -> None:
    a1 = any_gt(23)
    a2 = any_lt(42)
    a = a1 & a2 & ANY_INT
    assert isinstance(a, AnyAnd)
    assert len(a.args) == 3
    assert a.args[0] is a1
    assert a.args[1] is a2
    assert a.args[2] is ANY_INT


def test_and_and() -> None:
    a1 = any_gt(23)
    a2 = any_lt(42)
    a3 = ANY_INT
    a4 = any_func(lambda n: n % 2 == 1)
    a = (a1 & a2) & (a3 & a4)
    assert isinstance(a, AnyAnd)
    assert len(a.args) == 4
    assert a.args[0] is a1
    assert a.args[1] is a2
    assert a.args[2] is a3
    assert a.args[3] is a4
    assert a == 31
    assert a != 30
    assert a != 31.0
    assert a != 23
    assert a != 42



from typing import Any
import pytest
from anys import ANY_INT, ANY_STR, any_func, any_instance, not_
from test_lib import assert_equal, assert_not_equal


@pytest.mark.parametrize(
    "a,value",
    [
        (ANY_STR, 42),
        (ANY_INT, "foo"),
        (any_func(callable), "foo"),
        (any_instance((str, list)), 42),
        (42, 23),
        ("foo", "bar"),
        (None, 42),
    ],
)
def test_not_eq(a: Any, value: Any) -> None:
    assert_equal(not_(a), value)
    assert repr(not_(a)) == f"Not({a!r})"


@pytest.mark.parametrize(
    "a,value",
    [
        (ANY_STR, "foo"),
        (ANY_INT, 42),
        (any_func(callable), callable),
        (any_instance((str, list)), "foo"),
        (any_instance((str, list)), [1, 2, 3]),
        (42, 42),
        ("foo", "foo"),
        (None, None),
    ],
)
def test_not_neq(a: Any, value: Any) -> None:
    assert_not_equal(not_(a), value)



from collections.abc import Iterable, Mapping, Sequence
from datetime import date, datetime, time, timezone
from numbers import Number
from typing import Any
import pytest
from anys import (
    ANY_AWARE_DATETIME,
    ANY_AWARE_TIME,
    ANY_BOOL,
    ANY_BYTES,
    ANY_COMPLEX,
    ANY_DATE,
    ANY_DATETIME,
    ANY_DICT,
    ANY_FLOAT,
    ANY_INT,
    ANY_ITERABLE,
    ANY_ITERATOR,
    ANY_LIST,
    ANY_MAPPING,
    ANY_NAIVE_DATETIME,
    ANY_NAIVE_TIME,
    ANY_NUMBER,
    ANY_SEQUENCE,
    ANY_SET,
    ANY_STR,
    ANY_STRICT_DATE,
    ANY_TIME,
    ANY_TUPLE,
    ClassInfo,
    any_instance,
)
from test_lib import assert_equal, assert_not_equal


@pytest.mark.parametrize(
    "classinfo,value",
    [
        (str, "foo"),
        (int, 42),
        (tuple, ()),
        (list, [1, 2, 3]),
        ((str, list), "foo"),
        ((str, list), [1, 2, 3]),
    ],
)
def test_any_instance_eq(classinfo: ClassInfo, value: Any) -> None:
    assert_equal(any_instance(classinfo), value)
    assert repr(any_instance(classinfo)) == f"AnyInstance({classinfo!r})"


@pytest.mark.parametrize(
    "classinfo,value",
    [
        (str, 42),
        (int, "foo"),
        (int, 3.14),
        (tuple, []),
        (list, (1, 2, 3)),
        ((str, list), 42),
        ((str, list), (1, 2, 3)),
    ],
)
def test_any_instance_neq(classinfo: ClassInfo, value: Any) -> None:
    assert_not_equal(any_instance(classinfo), value)
    assert repr(any_instance(classinfo)) == f"AnyInstance({classinfo!r})"


@pytest.mark.parametrize("value", [True, False])
def test_any_bool_eq(value: bool) -> None:
    assert_equal(ANY_BOOL, value)


@pytest.mark.parametrize("value", [None, 42, "foo", (), 3.14, [], {}])
def test_any_bool_neq(value: Any) -> None:
    assert_not_equal(ANY_BOOL, value)


def test_any_bool_repr() -> None:
    assert repr(ANY_BOOL) == "ANY_BOOL"


@pytest.mark.parametrize("value", [b"", b"foo"])
def test_any_bytes_eq(value: bytes) -> None:
    assert_equal(ANY_BYTES, value)


@pytest.mark.parametrize("value", [None, 42, (), 3.14, [], {}, "foo"])
def test_any_bytes_neq(value: Any) -> None:
    assert_not_equal(ANY_BYTES, value)


def test_any_bytes_repr() -> None:
    assert repr(ANY_BYTES) == "ANY_BYTES"


@pytest.mark.parametrize("value", [1j, 1.2j, 1 + 2j, 1.2 + 3.4j])
def test_any_complex_eq(value: complex) -> None:
    assert_equal(ANY_COMPLEX, value)


@pytest.mark.parametrize("value", [None, "foo", (), 3.14, 42, [], {}])
def test_any_complex_neq(value: Any) -> None:
    assert_not_equal(ANY_COMPLEX, value)


def test_any_complex_repr() -> None:
    assert repr(ANY_COMPLEX) == "ANY_COMPLEX"


@pytest.mark.parametrize(
    "value",
    [
        date(2021, 6, 12),
        datetime(2021, 6, 12, 18, 3, 6),
        datetime(2021, 6, 12, 18, 3, 6, tzinfo=timezone.utc),
    ],
)
def test_any_date_eq(value: date) -> None:
    assert_equal(ANY_DATE, value)


@pytest.mark.parametrize(
    "value",
    [
        time(18, 3, 6),
        timezone.utc,
        None,
        42,
        "foo",
        (),
        3.14,
        [],
        {},
        b"bytes",
        "2021-06-12",
    ],
)
def test_any_date_neq(value: Any) -> None:
    assert_not_equal(ANY_DATE, value)


def test_any_date_repr() -> None:
    assert repr(ANY_DATE) == "ANY_DATE"


@pytest.mark.parametrize(
    "value",
    [
        datetime(2021, 6, 12, 18, 3, 6),
        datetime(2021, 6, 12, 18, 3, 6, tzinfo=timezone.utc),
    ],
)
def test_any_datetime_eq(value: datetime) -> None:
    assert_equal(ANY_DATETIME, value)


@pytest.mark.parametrize(
    "value",
    [
        date(2021, 6, 12),
        time(18, 3, 6),
        timezone.utc,
        None,
        42,
        "foo",
        (),
        3.14,
        [],
        {},
        b"bytes",
        "2021-06-12T18:03:06+00:00",
    ],
)
def test_any_datetime_neq(value: Any) -> None:
    assert_not_equal(ANY_DATETIME, value)


def test_any_datetime_repr() -> None:
    assert repr(ANY_DATETIME) == "ANY_DATETIME"


@pytest.mark.parametrize("value", [{}, {"foo": "bar"}])
def test_any_dict_eq(value: dict) -> None:
    assert_equal(ANY_DICT, value)


@pytest.mark.parametrize(
    "value", [None, 42, (), 3.14, [], "foo", set(), {1, 2}, b"bar"]
)
def test_any_dict_neq(value: Any) -> None:
    assert_not_equal(ANY_DICT, value)


def test_any_dict_repr() -> None:
    assert repr(ANY_DICT) == "ANY_DICT"


@pytest.mark.parametrize(
    "value", [-2.71828, 0.0, -0.0, 3.14, float("nan"), float("-inf"), float("inf")]
)
def test_any_float_eq(value: float) -> None:
    assert_equal(ANY_FLOAT, value)


@pytest.mark.parametrize("value", [None, "foo", (), 3, 1 + 2j, 1.2 + 3.4j, [], {}])
def test_any_float_neq(value: Any) -> None:
    assert_not_equal(ANY_FLOAT, value)


def test_any_float_repr() -> None:
    assert repr(ANY_FLOAT) == "ANY_FLOAT"


@pytest.mark.parametrize("value", [-23, 0, 42, True, False])
def test_any_int_eq(value: int) -> None:
    assert_equal(ANY_INT, value)


@pytest.mark.parametrize("value", [None, "foo", (), 3.14, 1 + 2j, 1.2 + 3.4j, [], {}])
def test_any_int_neq(value: Any) -> None:
    assert_not_equal(ANY_INT, value)


def test_any_int_repr() -> None:
    assert repr(ANY_INT) == "ANY_INT"


@pytest.mark.parametrize("value", ["foo", b"bar", (), [], {}, set(), iter([])])
def test_any_iterable_eq(value: Iterable) -> None:
    assert_equal(ANY_ITERABLE, value)


@pytest.mark.parametrize("value", [42, 3.14, 1 + 2j, None])
def test_any_iterable_neq(value: Any) -> None:
    assert_not_equal(ANY_ITERABLE, value)


def test_any_iterable_repr() -> None:
    assert repr(ANY_ITERABLE) == "ANY_ITERABLE"


def test_any_iterator_eq() -> None:
    assert_equal(ANY_ITERATOR, iter([]))


@pytest.mark.parametrize(
    "value", [42, 3.14, 1 + 2j, None, "foo", b"bar", (), [], {}, set()]
)
def test_any_iterator_neq(value: Any) -> None:
    assert_not_equal(ANY_ITERATOR, value)


def test_any_iterator_repr() -> None:
    assert repr(ANY_ITERATOR) == "ANY_ITERATOR"


@pytest.mark.parametrize("value", [[], [1], [1, 2, 3]])
def test_any_list_eq(value: list) -> None:
    assert_equal(ANY_LIST, value)


@pytest.mark.parametrize("value", [None, 42, "foo", 3.14, (), {}, b"bar"])
def test_any_list_neq(value: Any) -> None:
    assert_not_equal(ANY_LIST, value)


def test_any_list_repr() -> None:
    assert repr(ANY_LIST) == "ANY_LIST"


# TODO: Add non-dict mappings
@pytest.mark.parametrize("value", [{}, {"foo": "bar"}])
def test_any_mapping_eq(value: Mapping) -> None:
    assert_equal(ANY_MAPPING, value)


@pytest.mark.parametrize(
    "value", [None, 42, (), 3.14, [], "foo", set(), {1, 2}, b"bar"]
)
def test_any_mapping_neq(value: Any) -> None:
    assert_not_equal(ANY_MAPPING, value)


def test_any_mapping_repr() -> None:
    assert repr(ANY_MAPPING) == "ANY_MAPPING"


def test_any_naive_datetime_eq() -> None:
    assert_equal(ANY_NAIVE_DATETIME, datetime(2021, 6, 12, 18, 3, 6))


@pytest.mark.parametrize(
    "value",
    [
        datetime(2021, 6, 12, 18, 3, 6, tzinfo=timezone.utc),
        date(2021, 6, 12),
        time(18, 3, 6),
        timezone.utc,
        None,
        42,
        "foo",
        (),
        3.14,
        [],
        {},
        b"bytes",
        "2021-06-12T18:03:06",
    ],
)
def test_any_naive_datetime_neq(value: Any) -> None:
    assert_not_equal(ANY_NAIVE_DATETIME, value)


def test_any_naive_datetime_repr() -> None:
    assert repr(ANY_NAIVE_DATETIME) == "ANY_NAIVE_DATETIME"


def test_any_naive_time_eq() -> None:
    assert_equal(ANY_NAIVE_TIME, time(18, 3, 6))


@pytest.mark.parametrize(
    "value",
    [
        time(18, 3, 6, tzinfo=timezone.utc),
        date(2021, 6, 12),
        datetime(2021, 6, 12, 18, 3, 6),
        datetime(2021, 6, 12, 18, 3, 6, tzinfo=timezone.utc),
        timezone.utc,
        None,
        42,
        "foo",
        (),
        3.14,
        [],
        {},
        b"bytes",
        "18:03:06",
        "18:03:06+00:00",
    ],
)
def test_any_naive_time_neq(value: Any) -> None:
    assert_not_equal(ANY_NAIVE_TIME, value)


def test_any_naive_time_repr() -> None:
    assert repr(ANY_NAIVE_TIME) == "ANY_NAIVE_TIME"


@pytest.mark.parametrize(
    "value",
    [
        0,
        42,
        -23,
        True,
        False,
        -2.71828,
        0.0,
        -0.0,
        3.14,
        float("nan"),
        float("-inf"),
        float("inf"),
        1j,
        1.2j,
        1 + 2j,
        1.2 + 3.4j,
    ],
)
def test_any_number_eq(value: Number) -> None:
    assert_equal(ANY_NUMBER, value)


@pytest.mark.parametrize("value", [None, "foo", (), [], {}, b"bar"])
def test_any_number_neq(value: Any) -> None:
    assert_not_equal(ANY_NUMBER, value)


def test_any_number_repr() -> None:
    assert repr(ANY_NUMBER) == "ANY_NUMBER"


@pytest.mark.parametrize("value", ["foo", b"bar", (), []])
def test_any_sequence_eq(value: Sequence) -> None:
    assert_equal(ANY_SEQUENCE, value)


@pytest.mark.parametrize("value", [42, 3.14, 1 + 2j, None, {}, set(), iter([])])
def test_any_sequence_neq(value: Any) -> None:
    assert_not_equal(ANY_SEQUENCE, value)


def test_any_sequence_repr() -> None:
    assert repr(ANY_SEQUENCE) == "ANY_SEQUENCE"


@pytest.mark.parametrize("value", [set(), {1}, {1, 2, 3}])
def test_any_set_eq(value: set) -> None:
    assert_equal(ANY_SET, value)


@pytest.mark.parametrize("value", [None, 42, "foo", 3.14, (), [], {}, b"bar"])
def test_any_set_neq(value: Any) -> None:
    assert_not_equal(ANY_SET, value)


def test_any_set_repr() -> None:
    assert repr(ANY_SET) == "ANY_SET"


@pytest.mark.parametrize("value", ["", "foo"])
def test_any_str_eq(value: str) -> None:
    assert_equal(ANY_STR, value)


@pytest.mark.parametrize("value", [None, 42, (), 3.14, [], {}, b"bar"])
def test_any_str_neq(value: Any) -> None:
    assert_not_equal(ANY_STR, value)


def test_any_str_repr() -> None:
    assert repr(ANY_STR) == "ANY_STR"


def test_any_strict_date_eq() -> None:
    assert_equal(ANY_STRICT_DATE, date(2021, 6, 12))


@pytest.mark.parametrize(
    "value",
    [
        datetime(2021, 6, 12, 18, 3, 6),
        datetime(2021, 6, 12, 18, 3, 6, tzinfo=timezone.utc),
        time(18, 3, 6),
        timezone.utc,
        None,
        42,
        "foo",
        (),
        3.14,
        [],
        {},
        b"bytes",
        "2021-06-12",
    ],
)
def test_any_strict_date_neq(value: Any) -> None:
    assert_not_equal(ANY_STRICT_DATE, value)


def test_any_strict_date_repr() -> None:
    assert repr(ANY_STRICT_DATE) == "ANY_STRICT_DATE"


@pytest.mark.parametrize(
    "value",
    [
        time(18, 3, 6),
        time(18, 3, 6, tzinfo=timezone.utc),
    ],
)
def test_any_time_eq(value: time) -> None:
    assert_equal(ANY_TIME, value)


@pytest.mark.parametrize(
    "value",
    [
        date(2021, 6, 12),
        datetime(2021, 6, 12, 18, 3, 6),
        datetime(2021, 6, 12, 18, 3, 6, tzinfo=timezone.utc),
        timezone.utc,
        None,
        42,
        "foo",
        (),
        3.14,
        [],
        {},
        b"bytes",
        "18:03:06",
        "18:03:06+00:00",
    ],
)
def test_any_time_neq(value: Any) -> None:
    assert_not_equal(ANY_TIME, value)


def test_any_time_repr() -> None:
    assert repr(ANY_TIME) == "ANY_TIME"


@pytest.mark.parametrize("value", [(), (1,), (1, 2, 3)])
def test_any_tuple_eq(value: tuple) -> None:
    assert_equal(ANY_TUPLE, value)


@pytest.mark.parametrize("value", [None, 42, "foo", 3.14, [], {}, b"bar"])
def test_any_tuple_neq(value: Any) -> None:
    assert_not_equal(ANY_TUPLE, value)


def test_any_tuple_repr() -> None:
    assert repr(ANY_TUPLE) == "ANY_TUPLE"


def test_any_aware_datetime_eq() -> None:
    assert_equal(
        ANY_AWARE_DATETIME, datetime(2021, 6, 12, 18, 3, 6, tzinfo=timezone.utc)
    )


@pytest.mark.parametrize(
    "value",
    [
        datetime(2021, 6, 12, 18, 3, 6),
        date(2021, 6, 12),
        time(18, 3, 6),
        timezone.utc,
        None,
        42,
        "foo",
        (),
        3.14,
        [],
        {},
        b"bytes",
        "2021-06-12T18:03:06+00:00",
    ],
)
def test_any_aware_datetime_neq(value: Any) -> None:
    assert_not_equal(ANY_AWARE_DATETIME, value)


def test_any_aware_datetime_repr() -> None:
    assert repr(ANY_AWARE_DATETIME) == "ANY_AWARE_DATETIME"


def test_any_aware_time_eq() -> None:
    assert_equal(ANY_AWARE_TIME, time(18, 3, 6, tzinfo=timezone.utc))


@pytest.mark.parametrize(
    "value",
    [
        datetime(2021, 6, 12, 18, 3, 6),
        date(2021, 6, 12),
        time(18, 3, 6),
        timezone.utc,
        None,
        42,
        "foo",
        (),
        3.14,
        [],
        {},
        b"bytes",
        "2021-06-12T18:03:06+00:00",
    ],
)
def test_any_aware_time_neq(value: Any) -> None:
    assert_not_equal(ANY_AWARE_TIME, value)


def test_any_aware_time_repr() -> None:
    assert repr(ANY_AWARE_TIME) == "ANY_AWARE_TIME"



from typing import Any
import pytest
from anys import ANY_INT, any_ge, any_gt, any_le, any_lt
from test_lib import assert_equal, assert_not_equal


@pytest.mark.parametrize(
    "arg,value",
    [
        (42, 108),
        (42, 42),
        ("abc", "def"),
        ("abc", "abc"),
        ({"foo", 42}, {"foo", 42, 3.14}),
        ({"foo", 42}, {"foo", 42}),
    ],
)
def test_any_ge_eq(arg: Any, value: Any) -> None:
    assert_equal(any_ge(arg), value)
    assert repr(any_ge(arg)) == f"AnyGE({arg!r})"


@pytest.mark.parametrize(
    "arg,value",
    [
        (108, 42),
        ("def", "abc"),
        ({"foo", 42, 3.14}, {"foo", 42}),
        ({"foo"}, {42}),
        ({"foo", 42}, {42, 3.14}),
        ("1", 1),
        (ANY_INT, 23),
    ],
)
def test_any_ge_neq(arg: Any, value: Any) -> None:
    assert_not_equal(any_ge(arg), value)


@pytest.mark.parametrize(
    "arg,value",
    [
        (42, 108),
        ("abc", "def"),
        ({"foo", 42}, {"foo", 42, 3.14}),
    ],
)
def test_any_gt_eq(arg: Any, value: Any) -> None:
    assert_equal(any_gt(arg), value)
    assert repr(any_gt(arg)) == f"AnyGT({arg!r})"


@pytest.mark.parametrize(
    "arg,value",
    [
        (42, 42),
        ("abc", "abc"),
        ({"foo", 42}, {"foo", 42}),
        (108, 42),
        ("def", "abc"),
        ({"foo", 42, 3.14}, {"foo", 42}),
        ({"foo"}, {42}),
        ({"foo", 42}, {42, 3.14}),
        ("1", 1),
        (ANY_INT, 23),
    ],
)
def test_any_gt_neq(arg: Any, value: Any) -> None:
    assert_not_equal(any_gt(arg), value)


@pytest.mark.parametrize(
    "arg,value",
    [
        (108, 42),
        (42, 42),
        ("def", "abc"),
        ("abc", "abc"),
        ({"foo", 42, 31.4}, {"foo", 42}),
        ({"foo", 42}, {"foo", 42}),
    ],
)
def test_any_le_eq(arg: Any, value: Any) -> None:
    assert_equal(any_le(arg), value)
    assert repr(any_le(arg)) == f"AnyLE({arg!r})"


@pytest.mark.parametrize(
    "arg,value",
    [
        (42, 108),
        ("abc", "def"),
        ({"foo", 42}, {"foo", 42, 3.14}),
        ({"foo"}, {42}),
        ({"foo", 42}, {42, 3.14}),
        ("1", 1),
        (ANY_INT, 23),
    ],
)
def test_any_le_neq(arg: Any, value: Any) -> None:
    assert_not_equal(any_le(arg), value)


@pytest.mark.parametrize(
    "arg,value",
    [
        (108, 42),
        ("def", "abc"),
        ({"foo", 42, 3.14}, {"foo", 42}),
    ],
)
def test_any_lt_eq(arg: Any, value: Any) -> None:
    assert_equal(any_lt(arg), value)
    assert repr(any_lt(arg)) == f"AnyLT({arg!r})"


@pytest.mark.parametrize(
    "arg,value",
    [
        (42, 42),
        ("abc", "abc"),
        ({"foo", 42}, {"foo", 42}),
        (42, 108),
        ("abc", "def"),
        ({"foo", 42}, {"foo", 42, 3.14}),
        ({"foo"}, {42}),
        ({"foo", 42}, {42, 3.14}),
        ("1", 1),
        (ANY_INT, 23),
    ],
)
def test_any_lt_neq(arg: Any, value: Any) -> None:
    assert_not_equal(any_lt(arg), value)



"""
Matchers for pytest

``anys`` provides matchers for pytest_-style assertions.  What's a "matcher,"
you say?  Well, say you're writing a unit test and you want to assert that a
given object contains the correct values.  Normally, you'd just write:

.. code:: python

    assert foo == {
        "widgets": 42,
        "name": "Alice",
        "subfoo": {
            "created_at": "2021-06-24T18:41:59Z",
            "name": "Bob",
            "widgets": 23,
        }
    }

But wait!  What if the value of ``foo["subfoo"]["created_at"]`` can't be
determined in advance, but you still need to check that it's a valid timestamp?
You'd have to compare everything in ``foo`` other than that one field to the
expected values and then separately check the timestamp field for validity.
This is where matchers come in: they're magic objects that compare equal to any
& all values that meet given criteria.  For the case above, ``anys`` allows you
to just write:

.. code:: python

    from anys import ANY_DATETIME_STR

    assert foo == {
        "widgets": 42,
        "name": "Alice",
        "subfoo": {
            "created_at": ANY_DATETIME_STR,
            "name": "Bob",
            "widgets": 23,
        }
    }

and the assertion will do what you mean.

.. _pytest: https://docs.pytest.org

Visit <https://github.com/jwodder/anys> for more information.
"""

__version__ = "0.1.0"
__author__ = "John Thorvald Wodder II"
__author_email__ = "anys@varonathe.org"
__license__ = "MIT"
__url__ = "https://github.com/jwodder/anys"

from abc import ABC, abstractmethod
from collections import abc
from datetime import date, datetime, time
from numbers import Number
import operator
import re
import sys
import types
from typing import (
    Any,
    AnyStr,
    Callable,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
    Pattern,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar("T")

if sys.version_info[:2] >= (3, 10):
    ClassInfo = Union[
        type, types.Union, Tuple[Union[type, types.Union, Tuple[Any, ...]], ...]
    ]
else:
    ClassInfo = Union[type, Tuple[Union[type, Tuple[Any, ...]], ...]]

__all__ = [
    "ANY_AWARE_DATETIME",
    "ANY_AWARE_DATETIME_STR",
    "ANY_AWARE_TIME",
    "ANY_AWARE_TIME_STR",
    "ANY_BOOL",
    "ANY_BYTES",
    "ANY_COMPLEX",
    "ANY_DATE",
    "ANY_DATETIME",
    "ANY_DATETIME_STR",
    "ANY_DATE_STR",
    "ANY_DICT",
    "ANY_FALSY",
    "ANY_FLOAT",
    "ANY_INT",
    "ANY_ITERABLE",
    "ANY_ITERATOR",
    "ANY_LIST",
    "ANY_MAPPING",
    "ANY_NAIVE_DATETIME",
    "ANY_NAIVE_DATETIME_STR",
    "ANY_NAIVE_TIME",
    "ANY_NAIVE_TIME_STR",
    "ANY_NUMBER",
    "ANY_SEQUENCE",
    "ANY_SET",
    "ANY_STR",
    "ANY_STRICT_DATE",
    "ANY_TIME",
    "ANY_TIME_STR",
    "ANY_TRUTHY",
    "ANY_TUPLE",
    "any_contains",
    "any_fullmatch",
    "any_func",
    "any_ge",
    "any_gt",
    "any_in",
    "any_instance",
    "any_le",
    "any_lt",
    "any_match",
    "any_search",
    "any_substr",
    "any_with_attrs",
    "any_with_entries",
    "maybe",
    "not_",
]


class AnyBase(ABC):
    @abstractmethod
    def match(self, value: Any) -> bool:
        ...

    def __eq__(self, other: Any) -> bool:
        try:
            return self.match(other)
        except (TypeError, ValueError):
            return False

    def __and__(self, other: Any) -> Any:
        if isinstance(other, AnyBase):
            parts: List[AnyBase] = []
            for s in [self, other]:
                if isinstance(s, AnyAnd):
                    parts.extend(s.args)
                else:
                    parts.append(s)
            return AnyAnd(*parts)
        else:
            return NotImplemented  # pragma: no cover

    def __or__(self, other: Any) -> Any:
        if isinstance(other, AnyBase):
            parts: List[AnyBase] = []
            for s in [self, other]:
                if isinstance(s, AnyOr):
                    parts.extend(s.args)
                else:
                    parts.append(s)
            return AnyOr(*parts)
        else:
            return NotImplemented  # pragma: no cover


class AnyArg(AnyBase, Generic[T]):
    def __init__(self, arg: T, *, name: Optional[str] = None) -> None:
        self.arg = arg
        self.name = name

    def __repr__(self) -> str:
        if self.name is not None:
            return self.name
        else:
            return f"{type(self).__name__}({self.arg!r})"


class AnyFunc(AnyArg[Callable]):
    def match(self, value: Any) -> bool:
        return bool(self.arg(value))


# Anys need to be constructed via functions that return typing.Any so that
# comparing them against values with more restrictive types doesn't trigger a
# "comparison-overlap" type error in mypy.


def any_func(func: Callable) -> Any:
    """
    Returns a matcher that matches any value ``x`` for which ``func(x)`` is
    true.  If ``func(x)`` raises a `TypeError` or `ValueError`, it will be
    suppressed, and ``x == any_func(func)`` will evaluate to `False`.  All
    other exceptions are propagated out.
    """
    return AnyFunc(func)


class AnyInstance(AnyArg[ClassInfo]):
    def match(self, value: Any) -> bool:
        return isinstance(value, self.arg)


def any_instance(classinfo: ClassInfo, *, name: Optional[str] = None) -> Any:
    """
    Returns a matcher that matches any value that is an instance of
    ``classinfo``.  ``classinfo`` can be either a type or a tuple of types (or,
    starting in Python 3.10, a `Union` of types).
    """
    return AnyInstance(classinfo, name=name)


ANY_BOOL = any_instance(bool, name="ANY_BOOL")
ANY_BYTES = any_instance(bytes, name="ANY_BYTES")
ANY_COMPLEX = any_instance(complex, name="ANY_COMPLEX")
ANY_DATE = any_instance(date, name="ANY_DATE")
ANY_DATETIME = any_instance(datetime, name="ANY_DATETIME")
ANY_DICT = any_instance(dict, name="ANY_DICT")
ANY_FLOAT = any_instance(float, name="ANY_FLOAT")
ANY_INT = any_instance(int, name="ANY_INT")
ANY_ITERABLE = any_instance(abc.Iterable, name="ANY_ITERABLE")
ANY_ITERATOR = any_instance(abc.Iterator, name="ANY_ITERATOR")
ANY_LIST = any_instance(list, name="ANY_LIST")
ANY_MAPPING = any_instance(abc.Mapping, name="ANY_MAPPING")
ANY_NUMBER = any_instance(Number, name="ANY_NUMBER")
ANY_SEQUENCE = any_instance(abc.Sequence, name="ANY_SEQUENCE")
ANY_SET = any_instance(set, name="ANY_SET")
ANY_STR = any_instance(str, name="ANY_STR")
ANY_TIME = any_instance(time, name="ANY_TIME")
ANY_TUPLE = any_instance(tuple, name="ANY_TUPLE")


class AnyStrictDate(AnyBase):
    def match(self, value: Any) -> bool:
        return isinstance(value, date) and not isinstance(value, datetime)

    def __repr__(self) -> str:
        return "ANY_STRICT_DATE"


ANY_STRICT_DATE: Any = AnyStrictDate()


class AnyAwareDatetime(AnyBase):
    def match(self, value: Any) -> bool:
        return isinstance(value, datetime) and value.tzinfo is not None

    def __repr__(self) -> str:
        return "ANY_AWARE_DATETIME"


ANY_AWARE_DATETIME: Any = AnyAwareDatetime()


class AnyNaiveDatetime(AnyBase):
    def match(self, value: Any) -> bool:
        return isinstance(value, datetime) and value.tzinfo is None

    def __repr__(self) -> str:
        return "ANY_NAIVE_DATETIME"


ANY_NAIVE_DATETIME: Any = AnyNaiveDatetime()


class AnyAwareTime(AnyBase):
    def match(self, value: Any) -> bool:
        return isinstance(value, time) and value.tzinfo is not None

    def __repr__(self) -> str:
        return "ANY_AWARE_TIME"


ANY_AWARE_TIME: Any = AnyAwareTime()


class AnyNaiveTime(AnyBase):
    def match(self, value: Any) -> bool:
        return isinstance(value, time) and value.tzinfo is None

    def __repr__(self) -> str:
        return "ANY_NAIVE_TIME"


ANY_NAIVE_TIME: Any = AnyNaiveTime()


class Maybe(AnyArg[Any]):
    def match(self, value: Any) -> bool:
        return bool(value is None or self.arg == value)


def maybe(arg: Any) -> Any:
    """
    Returns a matcher that matches `None` and any value that equals or matches
    ``arg`` (which can be an `anys` matcher)
    """
    return Maybe(arg)


class Not(AnyArg[Any]):
    def match(self, value: Any) -> bool:
        return bool(self.arg != value)


def not_(arg: Any) -> Any:
    """
    Returns a matcher that matches anything that does not equal or match
    ``arg`` (which can be an `anys` matcher)
    """
    return Not(arg)


class AnyMatch(AnyArg[Union[AnyStr, Pattern[AnyStr]]]):
    def match(self, value: Any) -> bool:
        return bool(re.match(self.arg, value))


def any_match(patten: Union[AnyStr, Pattern[AnyStr]]) -> Any:
    """
    Returns a matcher that matches any string ``s`` for which
    ``re.match(pattern, s)`` succeeds
    """
    return AnyMatch(patten)


class AnySearch(AnyArg[Union[AnyStr, Pattern[AnyStr]]]):
    def match(self, value: Any) -> bool:
        return bool(re.search(self.arg, value))


def any_search(pattern: Union[AnyStr, Pattern[AnyStr]]) -> Any:
    """
    Returns a matcher that matches any string ``s`` for which
    ``re.search(pattern, s)`` succeeds
    """
    return AnySearch(pattern)


class AnyFullmatch(AnyArg[Union[AnyStr, Pattern[AnyStr]]]):
    def match(self, value: Any) -> bool:
        return bool(re.fullmatch(self.arg, value))


def any_fullmatch(
    pattern: Union[AnyStr, Pattern[AnyStr]], *, name: Optional[str] = None
) -> Any:
    """
    Returns a matcher that matches any string ``s`` for which
    ``re.fullmatch(pattern, s)`` succeeds
    """
    return AnyFullmatch(pattern, name=name)


class AnyIn(AnyArg[Iterable[T]]):
    def __init__(self, arg: Iterable[T], *, name: Optional[str] = None) -> None:
        self.arg: List[T] = list(arg)
        self.name = name

    def match(self, value: Any) -> bool:
        return bool(any(a == value for a in self.arg))


def any_in(iterable: Iterable) -> Any:
    """
    Returns a matcher that matches any value that equals or matches an element
    of ``iterable`` (which may contain `anys` matchers).  Note that, if
    ``iterable`` is a string, only individual characters in the string will
    match; to match substrings, use `any_substr()` instead.
    """
    return AnyIn(iterable)


class AnySubstr(AnyArg[AnyStr]):
    def match(self, value: Any) -> bool:
        return bool(value in self.arg)


def any_substr(s: AnyStr) -> Any:
    """Returns a matcher that matches any substring of ``s``"""
    return AnySubstr(s)


class AnyContains(AnyArg[Any]):
    def match(self, value: Any) -> bool:
        if isinstance(self.arg, AnyBase):
            return bool(any(self.arg == v for v in value))
        else:
            return bool(self.arg in value)


def any_contains(key: Any) -> Any:
    """
    Returns a matcher that matches any value for which ``key in value`` is
    true.  If ``key`` is an `anys` matcher, ``value == any_contains(key)``
    will instead be evaluated by iterating through the elements of ``value``
    and checking whether any match ``key``.
    """
    return AnyContains(key)


class AnyWithEntries(AnyArg[Mapping]):
    def match(self, value: Any) -> bool:
        for k, v in self.arg.items():
            try:
                if v != value[k]:
                    return False
            except LookupError:
                return False
        return True


def any_with_entries(mapping: Mapping) -> Any:
    """
    Returns a matcher that matches any object ``obj`` such that ``obj[k] == v``
    for all ``k,v`` in ``mapping.items()``.

    The values (but not the keys) of ``mapping`` can be `anys` matchers.
    """
    return AnyWithEntries(mapping)


class AnyWithAttrs(AnyArg[Mapping]):
    def match(self, value: Any) -> bool:
        for k, v in self.arg.items():
            try:
                if v != getattr(value, k):
                    return False
            except AttributeError:
                return False
        return True


def any_with_attrs(mapping: Mapping) -> Any:
    """
    Returns a matcher that matches any object ``obj`` such that ``getattr(obj,
    k) == v`` for all ``k,v`` in ``mapping.items()``.

    The values (but not the keys) of ``mapping`` can be `anys` matchers.
    """
    return AnyWithAttrs(mapping)


class AnyLT(AnyArg[Any]):
    def match(self, value: Any) -> bool:
        return bool(value < self.arg)


def any_lt(bound: Any) -> Any:
    """Returns a matcher that matches any value less than ``bound``"""
    return AnyLT(bound)


class AnyLE(AnyArg[Any]):
    def match(self, value: Any) -> bool:
        return bool(value <= self.arg)


def any_le(bound: Any) -> Any:
    """
    Returns a matcher that matches any value less than or equal to ``bound``
    """
    return AnyLE(bound)


class AnyGT(AnyArg[Any]):
    def match(self, value: Any) -> bool:
        return bool(value > self.arg)


def any_gt(bound: Any) -> Any:
    """Returns a matcher that matches any value greater than ``bound``"""
    return AnyGT(bound)


class AnyGE(AnyArg[Any]):
    def match(self, value: Any) -> bool:
        return bool(value >= self.arg)


def any_ge(bound: Any) -> Any:
    """
    Returns a matcher that matches any value greater than or equal to ``bound``
    """
    return AnyGE(bound)


ANY_TRUTHY = AnyFunc(bool, name="ANY_TRUTHY")
ANY_FALSY = AnyFunc(operator.not_, name="ANY_FALSY")


DATE_RGX = r"[0-9]{4,}-(?:0[1-9]|1[012])-(?:0[1-9]|[12][0-9]|3[01])"
TIME_RGX = r"(?:[01][0-9]|2[0-3]):[0-5][0-9](?::[0-5][0-9](?:\.[0-9]+)?)?"
TZ_RGX = r"(?:Z|[-+][0-9]{2}(?::?[0-9]{2})?)"

ANY_DATETIME_STR = any_fullmatch(
    re.compile(f"{DATE_RGX}[T ]{TIME_RGX}(?:{TZ_RGX})?"), name="ANY_DATETIME_STR"
)
ANY_AWARE_DATETIME_STR = any_fullmatch(
    re.compile(f"{DATE_RGX}[T ]{TIME_RGX}{TZ_RGX}"), name="ANY_AWARE_DATETIME_STR"
)
ANY_NAIVE_DATETIME_STR = any_fullmatch(
    re.compile(f"{DATE_RGX}[T ]{TIME_RGX}"), name="ANY_NAIVE_DATETIME_STR"
)

ANY_DATE_STR = any_fullmatch(re.compile(DATE_RGX), name="ANY_DATE_STR")

ANY_TIME_STR = any_fullmatch(
    re.compile(f"{TIME_RGX}(?:{TZ_RGX})?"), name="ANY_TIME_STR"
)
ANY_AWARE_TIME_STR = any_fullmatch(
    re.compile(f"{TIME_RGX}{TZ_RGX}"), name="ANY_AWARE_TIME_STR"
)
ANY_NAIVE_TIME_STR = any_fullmatch(re.compile(TIME_RGX), name="ANY_NAIVE_TIME_STR")


class AnyArgs(AnyBase):
    def __init__(self, *args: AnyBase, name: Optional[str] = None) -> None:
        self.args: List[AnyBase] = list(args)
        self.name = name

    def __repr__(self) -> str:
        return "{}({})".format(type(self).__name__, ", ".join(map(repr, self.args)))


class AnyAnd(AnyArgs):
    def match(self, value: Any) -> bool:
        return bool(all(a == value for a in self.args))


class AnyOr(AnyArgs):
    def match(self, value: Any) -> bool:
        return bool(any(a == value for a in self.args))
